-- **************************************************************************
--  This is the FIR filter for 4times over sampling
--       31,502kHz->128.008kHz
--			Fir  0-0.023:1   0.032-0.5:0/W:50  123*4taps  Q21
--				2013,12,28 by Y nishimura
-- **************************************************************************
LIBRARY ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
LIBRARY lpm;
use lpm.lpm_components.all;

entity Over4times is
	port
		(
			Hclk		: in  std_logic;	-- 65MHz clock
			Reset		: in  std_logic;	-- Master Reset
			Test		: in  std_logic;	-- Test signal for sim
			PDclk		: in  std_logic;	-- 1GHz/248=4.03225MHz(I/Q sampling)
			IdatIn		: in  std_logic_vector(17 downto 0);
			QdatIn		: in  std_logic_Vector(17 downto 0);
			Dgate		: out std_logic;	-- 4 times sampling gate
			Sample		: out std_logic;	-- Base sampling gate
			Hsamp		: out std_logic;	-- 4.03225MHz sampling gate
			Pulse		: out std_logic;
			Bstate		: out std_logic_vector(1 downto 0);  -- Base clock timing
			DaOutI		: out std_logic_vector(17 downto 0); -- Comb filter out
			DaOutQ		: out std_logic_vector(17 downto 0)  -- Comb filter out
		);
end Over4times;

architecture norm of Over4times is
	signal	MemoI,MemoO,Mulo	: std_logic_vector(35 downto 0);
	signal	Accl				: std_logic_vector(19 downto 0);
	signal	CoeffA,FirI,Acch	: std_logic_vector(17 downto 0);
	signal	Iout,Qout,Idao,Qdao	: std_logic_vector(17 downto 0);
	signal	Mulh				: std_logic_vector(16 downto 0);
	signal	Scount				: std_logic_vector(9 downto 0);
	signal	Romadd				: std_logic_vector(8 downto 0);
	signal	Wadd,Radd,Ocnt		: std_logic_vector(6 downto 0);
	signal	Ishape,Qshape		: std_logic_vector(6 downto 0);
	signal	Bcnt,Acnt			: std_logic_vector(4 downto 0);
	signal	Ssta,Csta,Fsta		: std_logic_vector(1 downto 0);
	signal	Gover,Bsamp,Dover	: std_logic;
	signal	Tdr,Gdla			: std_logic;
	signal	Dea,Deb,Dec,Ded		: std_logic;

begin
	Dgate <= Gdla; Sample <= Bsamp; Pulse <= Bcnt(3);
	DaOutI <= Iout; DaOutQ <= Qout; Bstate <= Ssta;
------------------------------------------------------------------------------------------
--		65MHz  countermesure for the Meta stable
--			1GHz/(62*4) = 4.32258MHz = 65MHz/16.12  0.12=3/25
------------------------------------------------------------------------------------------
	process(Hclk,Reset)
	begin
		if Reset = '0' 		then
			Bcnt <= "00000"; Acnt <= "00000"; Csta <= "00";
		elsif rising_edge(Hclk) then
			if Csta = "01"				then Acnt <= "00000"; Bcnt <= "00011";
			else
				if Bcnt = "01111"		then
					if Acnt = "11000"	then Acnt <= "00000";
										else Acnt <= Acnt + 1;
					end if;
					if Csta = "01"		then Bcnt <= "00000";
					else
						case Acnt is
							when "00111" => Bcnt <= Bcnt + 1;
							when "01111" => Bcnt <= Bcnt + 1;
							when "10111" => Bcnt <= Bcnt + 1;
							when others  => Bcnt <= "00000";
						end case;
					end if;
				else
					if Bcnt(4) = '1'	then Bcnt <= "00000";	
										else Bcnt <= Bcnt + 1;
					end if;
					Acnt <= Acnt;
				end if;
			end if;
			case Csta is
				when "00" => if PDclk = '1'		then
								if Bcnt(3 downto 2)  = "00"	then Csta <= "00";
															else Csta <= "01";
								end if;
							 else Csta <= "00";
							 end if;
				when "01" => Csta <= "11";
				when "10" => Csta <= "00";
				when "11" => if PDclk = '1'		then Csta <= "11";
												else Csta <= "10";
							 end if;
				when others => Csta <= "00";
			end case;
		end if;
	end process;

------------------------------------------------------------------------------------------
--		Timing GEN
------------------------------------------------------------------------------------------
--	I/Q clock(from AD9957) edge detecting ---
	process(Hclk,Reset)
	begin
		if Reset = '0' 		then Ssta <= "00";
		elsif rising_edge(Hclk) then
			case Ssta is
				when "00" => if Bcnt(3 downto 0) = "1100"	then Ssta <= "01";
															else Ssta <= "00";
							 end if;
				when "01" => Ssta <= "11";
				when "10" => Ssta <= "00";
				when "11" => if Bcnt(3) = '0'				then Ssta <= "10";
															else Ssta <= "11";
							 end if;
				when others => Ssta <= "00";
			end case;
		end if;
	end process;

-- 1/16 counter -----------------------------
	process(Hclk,Reset)
	begin
		if Reset = '0' 		then
			Ocnt <= "0000000"; Gover <= '0'; Bsamp <= '0'; Dover <= '0'; Hsamp <= '0';
			Gdla <= '0';
		elsif rising_edge(Hclk) then
			Gdla <= Gover;
			if Ssta = "01"			then Ocnt <= Ocnt + 1; Hsamp <= '1';
									else Ocnt <= Ocnt; Hsamp <= '0';
			end if;
			if Ssta = "01"		then
				if Ocnt(3 downto 0) = "0000"	then
					 Dover <= '1';
				else Dover <= '0';
				end if;
				if Ocnt(4 downto 0) = "00000"	then
					 Gover <= '1';
				else Gover <= '0';
				end if;
				if Ocnt = "0000000"	then Bsamp <= '1';
									else Bsamp <= '0';
				end if;
			else Gover <= '0'; Bsamp <= '0'; Dover <= '0'; 
			end if;
		end if;
	end process;

------------------------------------------------------------------------------------------
--		Status Counter
------------------------------------------------------------------------------------------
	process(Hclk,Reset)
	begin
		if Reset = '0' 		then Scount <= "0000000000";
		elsif rising_edge(Hclk) then
			if Bsamp = '1'	then Scount <= "0000000001";
			else
				if Scount(6 downto 0) = "0000000"	then
					if Dover = '1'	then Scount <= Scount + 1;
									else Scount <= Scount;
					end if;
				else Scount <= Scount + 1;
				end if;
			end if;
		end if;
	end process;

------------------------------------------------------------------------------------------
--		Data Buffering
------------------------------------------------------------------------------------------
-- Memory Read/Write address -----------------------------
	Tdr <= Bsamp or Test;
	process(Hclk,Reset)
	begin
		if Reset = '0' 		then Wadd <= "0000000"; Radd <= "0000000";
		elsif rising_edge(Hclk) then
--			if Bsamp = '1'	then Wadd <= Wadd - 1;
			if Tdr = '1'	then Wadd <= Wadd - 1;
							else Wadd <= Wadd;
			end if;
--			if Bsamp = '1'	then Radd <= Wadd;
			if Tdr = '1'	then Radd <= Wadd;
			else
				if Radd = Wadd	then
					if Dover = '1'	then Radd <= Radd + 1;
									else Radd <= Radd;
					end if;
				else Radd <= Radd + 1;
				end if;
			end if;
		end if;
	end process;

	MemoI(35 downto 18) <= IdatIn; MemoI(17 downto 0) <= QdatIn;
	U1: lpm_ram_dp
   		GENERIC map (
						LPM_WIDTH =>36, LPM_WIDTHAD => 7,
      					LPM_OUTDATA => "REGISTERED")
		port map (
					data => MemoI,
					wraddress => Wadd,
					wrclken => '1',
--					wren => Bsamp,
					wren => Tdr,
					wrclock => Hclk,
					rdaddress => Radd,
					rden => '1',
					rdclken => '1',
					rdclock => Hclk,
					q => MemoO
				  );

---------------------------------------------------------------------------------------
--		Over sampling FIR filter coefficient
---------------------------------------------------------------------------------------
	Romadd(8 downto 7) <= Scount(9 downto 8); Romadd(6 downto 0) <= Scount(6 downto 0);
	U2: LPM_ROM generic map(
				 LPM_WIDTH => 18, LPM_WIDTHAD => 9,
                 LPM_ADDRESS_CONTROL => "REGISTERED",
				 LPM_OUTDATA => "REGISTERED",
				 LPM_FILE => "FourTimesFir.MIF")
        port map(
					ADDRESS => Romadd,
                    INCLOCK => Hclk,
                    OUTCLOCK => Hclk,
					MEMENAB => '1',
					Q => CoeffA
				);

--------------------------------------------------------------------------------------
--		Data X FilterCoefficient
--------------------------------------------------------------------------------------
	process(MemoO,Scount)
	begin
		if Scount(7) = '1'	then FirI <= MemoO(17 downto 0);
							else FirI <= MemoO(35 downto 18);
		end if;
	end process;
	
	U3: lpm_mult GENERIC map (
			LPM_WIDTHA => 18,LPM_WIDTHB => 18,LPM_WIDTHP => 36,
			LPM_REPRESENTATION => "SIGNED",LPM_PIPELINE => 1) 
  	    PORT map (
					dataa => FirI,
					datab => CoeffA,
					aclr  => '0',
					clock => Hclk,
					clken => '1',
        			result => Mulo
				 ); 

--------------------------------------------------------------------------------------
--		Data Accumulation
--------------------------------------------------------------------------------------
 	process(Hclk,Reset)	  -- ACC calcu
		variable Tempa,Tempb,Tempc,Tempd : std_logic_vector(19 downto 0);

	begin
		if Reset = '0' then
			Accl <= "00000000000000000000"; Mulh <= "00000000000000000";
		elsif rising_edge(Hclk) then
			Tempa(19) := '0'; Tempb(19) := '0';
			Tempa(18 downto 0) := Accl(18 downto 0);
			Tempb(18 downto 0) := Mulo(18 downto 0);
			Mulh <= Mulo(35 downto 19);
			if Dover = '1' then
				Accl <= Tempb;
			else
				Accl <= Tempa + Tempb;	
			end if;
		end if;
	end process;

 	process(Hclk,Reset)	  -- ACC calcu
		variable Tempa : std_logic_vector(17 downto 0);
		
	begin
		if Reset = '0' then Acch <= "000000000000000000";
		elsif rising_edge(Hclk) then
			Tempa(17) := Mulh(16);
			Tempa(16 downto 0) := Mulh;
			if Dover = '1'	then Acch <= "000000000000000000";
			else
				if Accl(19) = '1'	then Acch <= Acch + Tempa + 1;
									else Acch <= Acch + Tempa;
				end if;
			end if;
		end if;
	end process;

--------------------------------------------------------------------------------------
--		Latch Data out
--------------------------------------------------------------------------------------
 	process(Hclk,Reset)	  -- ACC calcu
		variable Tempa,Tempb,Tempc,Tempd,Tempe : std_logic_vector(7 downto 0);
		variable Tempf,Temph : std_logic_vector(17 downto 0);
		
	begin
		if Reset = '0' then
			Iout <= "000000000000000000";
			Qout <= "000000000000000000";
			Ishape <= "0000000"; Qshape <= "0000000";
		elsif rising_edge(Hclk) then
			Tempa(7) := '0'; Tempa(6 downto 0) := Accl(17 downto 11);
			Tempb(7) := '0'; Tempb(6 downto 0) := Ishape;
			Tempc(7) := '0'; Tempc(6 downto 0) := Qshape;
			Tempd := Tempa + Tempb;
			Tempe := Tempa + Tempc;
			Tempf(17 downto 1) := Acch(16 downto 0); Tempf(0) := Accl(18);
			if Dover = '1'	then
				if Gover = '0'	then
					if Tempd(7) = '1'	then Iout <= Tempf + 1;
										else Iout <= Tempf;
					end if;
					Ishape <= Tempd(6 downto 0); Qshape <= Qshape;
				else
					if Tempe(7) = '1'	then Qout <= Tempf + 1;
										else Qout <= Tempf;
					end if;
					Ishape <= Ishape; Qshape <= Tempe(6 downto 0);
				end if;
			else
				Iout <= Iout; Qout <= Qout;
				Ishape <= Ishape; Qshape <= Qshape;
			end if;
		end if;
	end process;
	
end norm;
